/*
 * KernelConf - A Linux Kernel Configuration System
 *
 * Distributed under GNU General Public Lisence
 *
 * Copyright (C) 2002 by Anuradha Ratnaweera
 */

#include <ctype.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "readsymbols.h"
#include "symbols.h"

extern SymbolPtr symbol_tree_root;
extern SymbolPtr first_symbol;
extern MenuPtr menu_tree_root;
extern MenuPtr root_menu_item;

/*
 * static char *string_duplicate(char *str)
 *
 * Copies *str to malloc'd space after removing initial whitespaces
 */

static char *remove_leading_spaces(char *str)
{
    while ((*str) && isspace(*str))
	str++;
    return (str);
}

static char *string_duplicate(char *str)
{
    char *s;

    s = (char*)malloc(sizeof(char) * (strlen(str) + 1));
    strcpy(s, str);
    return s;
}

static char buffer[MAX_LINE_SIZE];

/*
 * static char *parse_line(FILE *symbfile, int &tag)
 *
 * Reads the next line from the config file and returns a pointer to
 * the value string in BUFFER.
 *
 * The tag type of the line is returned in TAG.
 */
static char *parse_line(FILE *symbfile, int *tag)
{
    int i;
    char *bufptr, c;

    i = 0;
    while (((c = fgetc(symbfile)) != EOF) && (c != '\n') && (i < MAX_LINE_SIZE))
	buffer[i++] = c;

    if ((c == EOF) && (i == 0)) {
	*tag = TAG_EOF;
	return NULL;
    }

    buffer[i] = 0;

    if (i == MAX_LINE_SIZE) {
	fprintf(stderr,
		"Warning: Line with length %d read, may have been truncated\n",
		MAX_LINE_SIZE);
    }

    /* remove comments */
    if ((bufptr = strchr(buffer, '#')))
	*bufptr = 0;
    else
	bufptr = buffer + i;

    /* bufptr now points to the end of the string */
    /* remove trailing whitespaces */
    while ((bufptr > buffer) && isspace(*(bufptr - 1)))
	bufptr--;
    *bufptr = 0;

    if (buffer[0] == 0) {
	*tag = TAG_BLANK;
	return NULL;
    }
    if (strncmp(buffer, "include", 7) == 0) {
	*tag = TAG_INCLUDE;
	return remove_leading_spaces(buffer + 7);
    }
    if (strncmp(buffer, "symbol:", 7) == 0) {
	*tag = TAG_SYMBOL;
	return remove_leading_spaces(buffer + 7);
    }
    if (strncmp(buffer, "type:", 5) == 0) {
	*tag = TAG_TYPE;
	return remove_leading_spaces(buffer + 5);
    }
    if (strncmp(buffer, "value:", 6) == 0) {
	*tag = TAG_VALUE;
	return remove_leading_spaces(buffer + 6);
    }
    if (strncmp(buffer, "description:", 12) == 0) {
	*tag = TAG_DESC;
	return NULL;
    }
    if (strncmp(buffer, "menu-item:", 10) == 0) {
	*tag = TAG_MENU_ITEM;
	return remove_leading_spaces(buffer + 10);
    }
    if (strncmp(buffer, "menu-caption:", 13) == 0) {
	*tag = TAG_MENU_CAPTION;
	return remove_leading_spaces(buffer + 13);
    }
    if (strncmp(buffer, "menu-parent:", 12) == 0) {
	*tag = TAG_MENU_PARENT;
	return remove_leading_spaces(buffer + 12);
    }

    /* is this a candidate for a description line? */
    if (isspace(buffer[0])) {
	bufptr = buffer;
	while ((*bufptr) && isspace(*bufptr))
	    bufptr++;
	if (*bufptr) {
	    *tag = TAG_DESC_LINE;
	    return remove_leading_spaces(bufptr);
	}
	else {
	    /* is this really necessary? */
	    *tag = TAG_BLANK;
	    return NULL;
	}
    }

    /* unknown tag, return NULL */
    *tag = TAG_NULL;
    return NULL;
}

void add_menu_to_parent(MenuPtr menu_item, MenuPtr parent)
{
    NodePtr ptr;
    if (parent->child) {
	ptr = (NodePtr)parent->child;
	while (ptr->next)
	    ptr = ptr->next;
	ptr->next = (NodePtr)menu_item;
    }
    else {
	parent->child = menu_item;
    }
}

/*
 * int readsymbols(char *filename)
 *
 * reads configuration symbols, menus and dependencies from .conf
 * files in the kernel tree */
int readsymbols(char *filename)
{
    int tag;
    FILE *symbfile;
    char *buf;
    int line_no;
    char last_line_blank;
    SymbolPtr symbol = NULL, last_symbol = NULL;
    MenuPtr menu_item = NULL;

#ifdef DEBUG
    printf("Processing file: %s\n", filename);
#endif

    if ((symbfile = fopen(filename, "rt")) == NULL) {
	fprintf(stderr, "Error: can't open %s\n", filename);
	exit(1);
    }

    line_no = 0;
    last_line_blank = 1;

    if (root_menu_item == NULL) {
	root_menu_item = (MenuPtr)search_tree((NodePtr*)(&menu_tree_root),
		"root", NODE_INSERT, sizeof(Menu), NULL);
	root_menu_item->parent = NULL;
	root_menu_item->child = NULL;
	((NodePtr)root_menu_item)->next = NULL;
	root_menu_item->caption = NULL;
	root_menu_item->depends = NULL;
	root_menu_item->symbol = NULL;
    }
	
    do {
	buf = parse_line(symbfile, &tag);
	line_no++;

	switch (tag) {
	    case TAG_NULL:
		fprintf(stderr, "Warning: %s:%d: unknown tag\n",
			filename, line_no);
		break;
	    case TAG_INCLUDE:
		readsymbols(buf);
		last_line_blank = 1;
		symbol = NULL;
		menu_item = NULL;
		break;
	    case TAG_SYMBOL:
		if (last_line_blank) {
		    last_line_blank = 0;
		    symbol = (SymbolPtr)search_tree((NodePtr*)(&symbol_tree_root),
			    buf, NODE_INSERT, sizeof(Symbol), (NodePtr)last_symbol);
		    last_symbol = symbol;
		    if (first_symbol == NULL)
			first_symbol = symbol;
		    symbol->type = TYPE_BOOL;
		    symbol->value.charval = 'n';
		    symbol->depends = NULL;
		    symbol->menu_item = NULL;
		}
		else {
		    fprintf(stderr,
			    "%s:%d: Symbol defined in the middle of the section\n",
			    filename, line_no);
		    exit(1);
		}
		break;
	    case TAG_TYPE:
		if (symbol == NULL) {
		    fprintf(stderr,
			    "%s:%d: Type definition without symbol\n",
			    filename, line_no);
		    exit(1);
		}
		if (strncmp(buf, "bool", 4) == 0) {
		    symbol->type = TYPE_BOOL;
		}
		else if (strncmp(buf, "trist", 5) == 0) {
		    symbol->type = TYPE_TRI;
		}
		else if (strncmp(buf, "int", 3) == 0) {
		    symbol->type = TYPE_INT;
		}
		else if (strncmp(buf, "char", 4) == 0) {
		    symbol->type = TYPE_CHAR;
		}
		else if (strncmp(buf, "str", 3) == 0) {
		    symbol->type = TYPE_STR;
		}
		break;
	    case TAG_VALUE:
		if (symbol == NULL) {
		    fprintf(stderr, "%s:%d: Value defined without symbol\n",
			    filename, line_no);
		    exit(1);
		}
		switch (symbol->type) {
		    case TYPE_BOOL:
			if ((buf[0] == 'Y') || (buf[0] == 'y'))
			    symbol->value.charval = 'y';
			else
			    symbol->value.charval = 'n';
			break;
		    case TYPE_TRI:
			if ((buf[0] == 'Y') || (buf[0] == 'y'))
			    symbol->value.charval = 'y';
			else if ((buf[0] == 'M') || (buf[0] == 'm'))
			    symbol->value.charval = 'm';
			else
			    symbol->value.charval = 'n';
			break;
		    case TYPE_INT:
			symbol->value.intval = atoi(buf);
			break;
		    case TYPE_CHAR:
			symbol->value.intval = buf[0];
			break;
		    case TYPE_STR:
			symbol->value.strval = buf;
			break;
		    case TYPE_NULL:
			break;
		}
		break;
	    case TAG_MENU_ITEM:
		menu_item = (MenuPtr)search_tree((NodePtr*)(&menu_tree_root),
			buf, NODE_INSERT, sizeof(Menu), NULL);
		menu_item->parent = NULL;
		menu_item->child = NULL;
		((NodePtr)menu_item)->next = NULL;
		menu_item->caption = NULL;
		menu_item->depends = NULL;
		if (last_line_blank)
		    last_line_blank = 0;
		if (symbol) {
		    menu_item->symbol = symbol;
		    symbol->menu_item = menu_item;
		}
		else {
		    menu_item->symbol = NULL;
		}
		break;
	    case TAG_MENU_CAPTION:
		if (menu_item == NULL) {
		    fprintf(stderr,
			    "%s:%d: Menu-caption defined without menu item\n",
			    filename, line_no);
		    exit(1);
		}
		menu_item->caption = string_duplicate(buf);
		break;
	    case TAG_MENU_PARENT:
		if (menu_item == NULL) {
		    fprintf(stderr,
			    "%s:%d: Menu-parent defined without menu item\n",
			    filename, line_no);
		    exit(1);
		}
		menu_item->parent = (MenuPtr)search_tree((NodePtr*)(&menu_tree_root),
			buf, NODE_MARK_INSERT, sizeof(Menu), NULL);
		add_menu_to_parent(menu_item, menu_item->parent);
		break;
	    case TAG_BLANK:
		last_line_blank = 1;
		symbol = NULL;
	    case TAG_EOF:
		if (menu_item && (menu_item->parent == NULL)) {
		    add_menu_to_parent(menu_item, root_menu_item);
		    menu_item->parent = root_menu_item;
		}
		menu_item = NULL;
		break;
	    default:
		;
	}

    } while (tag != TAG_EOF);



    fclose(symbfile);

    return 0;
}

